"""tests for vak.cli.predict module"""
from unittest import mock
import pytest

import vak.cli.predict
import vak.config
import vak.constants
import vak.paths

from . import cli_asserts


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
    ],
)
def test_predict(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    output_dir = tmp_path.joinpath(
        f"test_predict_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "PREDICT", "option": "output_dir", "value": str(output_dir)},
        {"section": "PREDICT", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="predict",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        options_to_change=options_to_change,
    )

    with mock.patch('vak.core.predict', autospec=True) as mock_core_predict:
        vak.cli.predict.predict(toml_path)
        assert mock_core_predict.called

    assert cli_asserts.log_file_created(command="predict", output_path=output_dir)
    assert cli_asserts.log_file_contains_version(command="predict", output_path=output_dir)


def test_predict_csv_path_none_raises(
        specific_config, tmp_path,
):
    """Test that cli.predict raises ValueError when csv_path is None
    (presumably because `vak prep` was not run yet)
    """
    options_to_change = [
        {"section": "PREDICT", "option": "csv_path", "value": "DELETE-OPTION"},
    ]

    toml_path = specific_config(
        config_type="predict",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )

    with pytest.raises(ValueError):
        vak.cli.predict.predict(toml_path)



"""tests for vak.core.train module"""
import pytest

import vak.config
import vak.constants
import vak.paths
import vak.core.train


def train_output_matches_expected(cfg, model_config_map, results_path):
    assert results_path.joinpath("labelmap.json").exists()

    if cfg.train.normalize_spectrograms or cfg.train.spect_scaler_path:
        assert results_path.joinpath("StandardizeSpect").exists()
    else:
        assert not results_path.joinpath("StandardizeSpect").exists()

    for model_name in model_config_map.keys():
        model_path = results_path.joinpath(model_name)
        assert model_path.exists()

        tensorboard_log = sorted(model_path.glob(f"events.out.tfevents.*{model_name}"))
        assert len(tensorboard_log) == 1

        checkpoints_path = model_path.joinpath("checkpoints")
        assert checkpoints_path.exists()
        assert checkpoints_path.joinpath("checkpoint.pt").exists()
        if cfg.train.val_step is not None:
            assert checkpoints_path.joinpath("max-val-acc-checkpoint.pt").exists()

    return True


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_train(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()
    options_to_change = [
        {"section": "TRAIN", "option": "device", "value": device},
        {"section": "TRAIN", "option": "root_results_dir", "value": results_path}
    ]
    toml_path = specific_config(
        config_type="train",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)

    vak.core.train(
        model_config_map,
        cfg.train.csv_path,
        cfg.dataloader.window_size,
        cfg.train.batch_size,
        cfg.train.num_epochs,
        cfg.train.num_workers,
        labelset=cfg.prep.labelset,
        results_path=results_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.train.normalize_spectrograms,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
        device=cfg.train.device,
    )

    assert train_output_matches_expected(cfg, model_config_map, results_path)


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_continue_training(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()
    options_to_change = [
        {"section": "TRAIN", "option": "device", "value": device},
        {"section": "TRAIN", "option": "root_results_dir", "value": results_path}
    ]
    toml_path = specific_config(
        config_type="train_continue",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)

    vak.core.train(
        model_config_map=model_config_map,
        csv_path=cfg.train.csv_path,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.train.batch_size,
        num_epochs=cfg.train.num_epochs,
        num_workers=cfg.train.num_workers,
        labelmap_path=cfg.train.labelmap_path,
        spect_scaler_path=cfg.train.spect_scaler_path,
        results_path=results_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.train.normalize_spectrograms,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
        device=cfg.train.device,
    )

    assert train_output_matches_expected(cfg, model_config_map, results_path)


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"section": "TRAIN", "option": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
        {"section": "TRAIN", "option": "labelmap_path", "value": '/obviously/doesnt/exist/labelmap.json'},
        {"section": "TRAIN", "option": "csv_path", "value": '/obviously/doesnt/exist/dataset.csv'},
        {"section": "TRAIN", "option": "spect_scaler_path", "value": '/obviously/doesnt/exist/SpectScaler'},
    ]
)
def test_train_raises_file_not_found(
    path_option_to_change, specific_config, tmp_path, device
):
    """Test that pre-conditions in `vak.core.train` raise FileNotFoundError
    when one of the following does not exist:
    checkpoint_path, labelmap_path, csv_path, spect_scaler_path
    """
    options_to_change = [
        {"section": "TRAIN", "option": "device", "value": device},
        path_option_to_change
    ]
    toml_path = specific_config(
        config_type="train",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    with pytest.raises(FileNotFoundError):
        vak.core.train(
            model_config_map=model_config_map,
            csv_path=cfg.train.csv_path,
            window_size=cfg.dataloader.window_size,
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            labelset=cfg.prep.labelset,
            labelmap_path=cfg.train.labelmap_path,
            checkpoint_path=cfg.train.checkpoint_path,
            spect_scaler_path=cfg.train.spect_scaler_path,
            results_path=results_path,
            spect_key=cfg.spect_params.spect_key,
            timebins_key=cfg.spect_params.timebins_key,
            normalize_spectrograms=cfg.train.normalize_spectrograms,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
            patience=cfg.train.patience,
            device=cfg.train.device,
        )


def test_train_raises_not_a_directory(
    specific_config, device
):
    """Test that core.train raises NotADirectory
    when ``results_path`` does not exist
    """
    options_to_change = [
        {"section": "TRAIN", "option": "root_results_dir", "value": '/obviously/doesnt/exist/results/'},
        {"section": "TRAIN", "option": "device", "value": device},
    ]
    toml_path = specific_config(
        config_type="train",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)

    # mock behavior of cli.train, building `results_path` from config option `root_results_dir`
    results_path = cfg.train.root_results_dir / 'results-dir-timestamp'

    with pytest.raises(NotADirectoryError):
        vak.core.train(
            model_config_map=model_config_map,
            csv_path=cfg.train.csv_path,
            window_size=cfg.dataloader.window_size,
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            labelset=cfg.prep.labelset,
            labelmap_path=cfg.train.labelmap_path,
            checkpoint_path=cfg.train.checkpoint_path,
            spect_scaler_path=cfg.train.spect_scaler_path,
            results_path=results_path,
            spect_key=cfg.spect_params.spect_key,
            timebins_key=cfg.spect_params.timebins_key,
            normalize_spectrograms=cfg.train.normalize_spectrograms,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
            patience=cfg.train.patience,
            device=cfg.train.device,
        )


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_both_labelset_and_labelmap_raises(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    """Test that ``core.train`` raises a ValueError when
    we pass in arguments for both the ``labelset`` parameter
    and the ``labelmap_path`` parameter.
    """
    options_to_change = {"section": "TRAIN", "option": "device", "value": device}
    toml_path = specific_config(
        config_type="train_continue",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)
    results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    with pytest.raises(ValueError):
        vak.core.train(
            model_config_map=model_config_map,
            csv_path=cfg.train.csv_path,
            window_size=cfg.dataloader.window_size,
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            labelset=cfg.prep.labelset,
            labelmap_path=cfg.train.labelmap_path,
            spect_scaler_path=cfg.train.spect_scaler_path,
            results_path=results_path,
            spect_key=cfg.spect_params.spect_key,
            timebins_key=cfg.spect_params.timebins_key,
            normalize_spectrograms=cfg.train.normalize_spectrograms,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
            patience=cfg.train.patience,
            device=cfg.train.device,
        )



import subprocess


def capture(command):
    result = subprocess.run(command, capture_output=True, text=True)
    return result.stdout, result.stderr, result.returncode


def test_cli_just_vak():
    command = ["vak"]
    out, err, returncode = capture(command)
    assert returncode > 0
    assert out == ''
    assert err.startswith('usage: vak ')


def test_cli_help():
    command = ["vak", "-h"]
    out, err, returncode = capture(command)
    assert returncode == 0
    assert out.startswith('usage: vak ')
    assert err == ''



import pandas as pd
import pytest


@pytest.fixture
def specific_dataframe(specific_prep_csv_path):
    """returns a function that will return a
    dataframe corresponding to the ``prep``ared dataset
    from a specific configuration file,
    determined by characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`
    """

    def _specific_dataframe(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
    ):
        csv_path = specific_prep_csv_path(
            config_type, model, annot_format, audio_format, spect_format
        )
        return pd.read_csv(csv_path)

    return _specific_dataframe


@pytest.fixture
def train_cbin_notmat_df(specific_dataframe):
    """Returns a specific dataframe
    for tests that don't need to use a factory,
    they just need some dataframe
    representing a dataset to test on,
    e.g., the ``SpectStandardize.fit_df`` method
    """
    return specific_dataframe(
        config_type="train",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat"
    )



"""tests for vak.config.parse module"""
import copy

import pytest

import vak.config
import vak.transforms.transforms
import vak.split
import vak.models
import vak.spect


@pytest.mark.parametrize(
    "section_name",
    [
        "DATALOADER",
        "EVAL" "LEARNCURVE",
        "PREDICT",
        "PREP",
        "SPECT_PARAMS",
        "TRAIN",
    ],
)
def test_parse_config_section_returns_attrs_class(
    section_name,
    configs_toml_path_pairs_by_model_factory,
    model,
):
    """test that ``vak.config.parse.parse_config_section``
    returns an instance of ``vak.config.learncurve.LearncurveConfig``"""
    config_toml_path_pairs = configs_toml_path_pairs_by_model_factory(model, section_name)
    for config_toml, toml_path in config_toml_path_pairs:
        config_section_obj = vak.config.parse.parse_config_section(
            config_toml=config_toml,
            section_name=section_name,
            toml_path=toml_path,
        )
        assert isinstance(
            config_section_obj, vak.config.parse.SECTION_CLASSES[section_name]
        )


@pytest.mark.parametrize(
    "section_name",
    [
        "DATALOADER",
        "EVAL",
        "LEARNCURVE",
        "PREDICT",
        "PREP",
        "SPECT_PARAMS",
        "TRAIN",
    ],
)
def test_parse_config_section_missing_options_raises(
    section_name,
    configs_toml_path_pairs_by_model_factory,
    model,
):
    """test that configs without the required options in a section raise KeyError"""
    if vak.config.parse.REQUIRED_OPTIONS[section_name] is None:
        pytest.skip(f"no required options to test for section: {section_name}")

    configs_toml_path_pairs = configs_toml_path_pairs_by_model_factory(model, section_name)

    for config_toml, toml_path in configs_toml_path_pairs:
        for option in vak.config.parse.REQUIRED_OPTIONS[section_name]:
            config_copy = copy.deepcopy(config_toml)
            config_copy[section_name].pop(option)
            with pytest.raises(KeyError):
                config = vak.config.parse.parse_config_section(
                    config_toml=config_copy,
                    section_name=section_name,
                    toml_path=toml_path,
                )


@pytest.mark.parametrize("section_name", ["EVAL", "LEARNCURVE", "PREDICT", "TRAIN"])
def test_parse_config_section_model_not_installed_raises(
        section_name,
        configs_toml_path_pairs_by_model_factory,
        model,
):
    """test that a ValueError is raised when the ``models`` option
    in the section specifies names of models that are not installed"""
    # we only need one toml, path pair
    # so we just call next on the ``zipped`` iterator that our fixture gives us
    configs_toml_path_pairs = configs_toml_path_pairs_by_model_factory(model)

    for config_toml, toml_path in configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these. Only need to test on one

    config_toml[section_name]["models"] = "NotInstalledNet, OtherNotInstalledNet"
    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(
            config_toml=config_toml, section_name=section_name, toml_path=toml_path
        )


def test_parse_prep_section_both_audio_and_spect_format_raises(
    all_generated_configs_toml_path_pairs,
):
    """test that a config with both an audio and a spect format raises a ValueError"""
    # iterate through configs til we find one with an `audio_format` option
    # and then we'll add a `spect_format` option to it
    found_config_to_use = False
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if "audio_format" in config_toml["PREP"]:
            found_config_to_use = True
            break
    assert found_config_to_use  # sanity check

    config_toml["PREP"]["spect_format"] = "mat"
    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(config_toml, "PREP", toml_path)


def test_parse_prep_section_neither_audio_nor_spect_format_raises(
    all_generated_configs_toml_path_pairs,
):
    """test that a config without either an audio or a spect format raises a ValueError"""
    # iterate through configs til we find one with an `audio_format` option
    # and then we'll add a `spect_format` option to it
    found_config_to_use = False
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if "audio_format" in config_toml["PREP"]:
            found_config_to_use = True
            break
    assert found_config_to_use  # sanity check

    config_toml["PREP"].pop("audio_format")
    if "spect_format" in config_toml["PREP"]:
        # shouldn't be but humor me
        config_toml["PREP"].pop("spect_format")

    with pytest.raises(ValueError):
        vak.config.parse.parse_config_section(config_toml, "PREP", toml_path)


def test_load_from_toml_path(all_generated_configs):
    for toml_path in all_generated_configs:
        config_toml = vak.config.parse._load_toml_from_path(toml_path)
        assert isinstance(config_toml, dict)


def test_load_from_toml_path_raises_when_config_doesnt_exist(config_that_doesnt_exist):
    with pytest.raises(FileNotFoundError):
        vak.config.parse._load_toml_from_path(config_that_doesnt_exist)


def test_from_toml_path_returns_instance_of_config(
    all_generated_configs, default_model
):
    for toml_path in all_generated_configs:
        if default_model not in str(toml_path):
            continue  # only need to check configs for one model
            # also avoids FileNotFoundError on CI
        config_obj = vak.config.parse.from_toml_path(toml_path)
        assert isinstance(config_obj, vak.config.parse.Config)


def test_from_toml_path_raises_when_config_doesnt_exist(config_that_doesnt_exist):
    with pytest.raises(FileNotFoundError):
        vak.config.parse.from_toml_path(config_that_doesnt_exist)


def test_from_toml(configs_toml_path_pairs_by_model_factory, model):
    config_toml_path_pairs = configs_toml_path_pairs_by_model_factory(model)
    for config_toml, toml_path in config_toml_path_pairs:
        config_obj = vak.config.parse.from_toml(config_toml, toml_path)
        assert isinstance(config_obj, vak.config.parse.Config)


def test_from_toml_parse_prep_with_sections_not_none(
    configs_toml_path_pairs_by_model_factory,
    model,
):
    """test that we get only the sections we want when we pass in a sections list to
    ``from_toml``. Specifically test ``PREP`` since that's what this will be used for."""
    # only use configs from 'default_model') (teenytweetynet)
    # so we are sure paths exist, to avoid NotADirectoryErrors that give spurious test failures
    config_toml_path_pairs = configs_toml_path_pairs_by_model_factory(model)
    for config_toml, toml_path in config_toml_path_pairs:
        config_obj = vak.config.parse.from_toml(
            config_toml, toml_path, sections=["PREP", "SPECT_PARAMS"]
        )
        assert isinstance(config_obj, vak.config.parse.Config)
        for should_have in ("prep", "spect_params"):
            assert hasattr(config_obj, should_have)
        for should_be_none in ("eval", "learncurve", "train", "predict"):
            assert getattr(config_obj, should_be_none) is None
        assert (
            getattr(config_obj, "dataloader")
            == vak.config.dataloader.DataLoaderConfig()
        )


@pytest.mark.parametrize("section_name", ["EVAL", "LEARNCURVE", "PREDICT", "TRAIN"])
def test_from_toml_parse_prep_with_sections_not_none(
    section_name, all_generated_configs_toml_path_pairs, random_path_factory
):
    """Test that ``config.parse.from_toml`` parameter ``sections`` works as expected.

    If we pass in a list of section names
    specifying that we only want to parse ``PREP`` and ``SPECT_PARAMS``,
    other sections should be left as None in the return Config instance."""
    for config_toml, toml_path in all_generated_configs_toml_path_pairs:
        if section_name.lower() in toml_path.name:
            break  # use these

    purpose = vak.cli.prep.purpose_from_toml(config_toml, toml_path)
    section_name = purpose.upper()
    required_options = vak.config.parse.REQUIRED_OPTIONS[section_name]
    for required_option in required_options:
        # set option to values that **would** cause an error if we parse them
        if "path" in required_option:
            badval = random_path_factory(f"_{required_option}.exe")
        elif "dir" in required_option:
            badval = random_path_factory("nonexistent_dir")
        else:
            continue
        config_toml[section_name][required_option] = badval
    cfg = vak.config.parse.from_toml(
        config_toml, toml_path, sections=["PREP", "SPECT_PARAMS"]
    )
    assert hasattr(cfg, 'prep') and getattr(cfg, 'prep') is not None
    assert hasattr(cfg, 'spect_params') and getattr(cfg, 'spect_params') is not None
    assert getattr(cfg, purpose) is None


def test_invalid_section_raises(invalid_section_config_path):
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_section_config_path)


def test_invalid_option_raises(invalid_option_config_path):
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_option_config_path)


@pytest.fixture
def invalid_train_and_learncurve_config_toml(test_configs_root):
    return test_configs_root.joinpath("invalid_train_and_learncurve_config.toml")


def test_train_and_learncurve_defined_raises(invalid_train_and_learncurve_config_toml):
    """test that a .toml config with both a TRAIN and a LEARNCURVE section raises a ValueError"""
    with pytest.raises(ValueError):
        vak.config.parse.from_toml_path(invalid_train_and_learncurve_config_toml)



import pytest

import vak


@pytest.mark.parametrize(
    'config_type, audio_format, spect_format, annot_format, expected_result',
    [
        ("train", "cbin", None, "notmat", True),
        ("train", "wav", None, "birdsong-recognition-dataset", True),
        ("train", None, "mat", "yarden", True),
    ]
)
def test_has_unlabeled(config_type,
                       audio_format,
                       spect_format,
                       annot_format,
                       expected_result,
                       model,
                       specific_config_toml,
                       specific_prep_csv_path):
    csv_path = specific_prep_csv_path(config_type,
                                      model,
                                      annot_format,
                                      audio_format,
                                      spect_format)

    has_unlabeled = vak.datasets.seq.validators.has_unlabeled(csv_path)
    assert has_unlabeled == expected_result



import crowsetta
import evfuncs
import numpy as np
import pytest

import vak.annotation
import vak.io.audio


def test_files_from_dir(annot_dir_notmat, annot_files_notmat):
    annot_files_from_dir = vak.annotation.files_from_dir(
        annot_dir_notmat, annot_format="notmat"
    )

    annot_files_notmat = [str(annot_file) for annot_file in annot_files_notmat]
    assert sorted(annot_files_from_dir) == sorted(annot_files_notmat)


@pytest.mark.parametrize(
    'path, audio_ext, expected_stem',
    [
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin.not.mat', None, 'gy6or6_baseline_230312_0808.138'),
        ('`/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin.not.mat', 'cbin', 'gy6or6_baseline_230312_0808.138'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin.not.mat', '.cbin', 'gy6or6_baseline_230312_0808.138'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin', None, 'gy6or6_baseline_230312_0808.138'),
        ('`/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin', 'cbin', 'gy6or6_baseline_230312_0808.138'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.cbin', '.cbin', 'gy6or6_baseline_230312_0808.138'),
        ('gy6or6_baseline_230312_0808.138.cbin.not.mat', None, 'gy6or6_baseline_230312_0808.138'),
        ('gy6or6_baseline_230312_0808.138.cbin.not.mat', 'cbin', 'gy6or6_baseline_230312_0808.138'),
        ('gy6or6_baseline_230312_0808.138.cbin.not.mat', '.cbin', 'gy6or6_baseline_230312_0808.138'),
        ('Bird0/spectrograms/0.wav.npz', None, '0'),
        ('Bird0/spectrograms/0.wav.npz', 'wav', '0'),
        ('Bird0/spectrograms/0.wav.npz', '.wav', '0'),
        ('Bird0/spectrograms/0.wav', None, '0'),
        ('0.wav', None, '0'),
        ('0.wav', 'wav', '0'),
        ('0.wav', '.wav', '0'),
    ]
)
def test_audio_stem_from_path(path, audio_ext, expected_stem):
    stem = vak.annotation.audio_stem_from_path(path, audio_ext)
    assert stem == expected_stem


@pytest.mark.parametrize(
    'path, audio_ext',
    [
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.', None),
        ('`/gy6or6/032212/gy6or6_baseline_230312_0808.138.', 'cbin'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.', '.cbin'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.', None),
        ('`/gy6or6/032212/gy6or6_baseline_230312_0808.138.', 'cbin'),
        ('~/gy6or6/032212/gy6or6_baseline_230312_0808.138.', '.cbin'),
        ('gy6or6_baseline_230312_0808.138.', None),
        ('gy6or6_baseline_230312_0808.138.', 'cbin'),
        ('gy6or6_baseline_230312_0808.138.', '.cbin'),
        ('Bird0/spectrograms/0', None,),
        ('Bird0/spectrograms/0.', 'wav'),
        ('Bird0/spectrograms/0.', '.wav'),
        ('Bird0/spectrograms/0.', None),
        ('0.', None),
        ('0.', 'wav'),
        ('0.', '.wav'),
    ]
)
def test_audio_stem_from_path_raises(path, audio_ext):
    with pytest.raises(vak.annotation.AudioFilenameNotFound):
        vak.annotation.audio_stem_from_path(path, audio_ext)


@pytest.mark.parametrize(
    "source_type, source_format, annot_format, audio_ext",
    [
        ("audio", "cbin", "notmat", None),
        ("audio", "wav", "birdsong-recognition-dataset", None),
        ("spect", "mat", "yarden", None),
        ("audio", "cbin", "notmat", "cbin"),
        ("audio", "wav", "birdsong-recognition-dataset", "wav"),
        ("spect", "mat", "yarden", "wav"),
    ],
)
def test__map_using_audio_stem_from_path(
    source_type,
    source_format,
    annot_format,
    audio_ext,
    audio_list_factory,
    spect_list_mat,
    specific_annot_list,
):
    if source_type == "audio":
        annotated_files = audio_list_factory(source_format, annot_format)
    else:
        annotated_files = spect_list_mat
    annot_list = specific_annot_list(annot_format)

    annotated_annot_map = vak.annotation._map_using_audio_stem_from_path(
        annotated_files=annotated_files, annot_list=annot_list, audio_ext=audio_ext
    )

    # test all the audio paths made it into the map
    annotated_files_from_map = list(annotated_annot_map.keys())
    for source_file in annotated_files:
        assert source_file in annotated_files_from_map

    # test all the annots made it into the map
    annot_list_from_map = list(annotated_annot_map.values())
    for annot in annot_list:
        assert annot in annot_list_from_map

    # test all mappings are correct
    for source_path, annot in list(annotated_annot_map.items()):
        assert vak.annotation.audio_stem_from_path(
            annot.audio_path
        ) == vak.annotation.audio_stem_from_path(source_path)


@pytest.mark.parametrize(
    "source_type, source_format, annot_format, annotated_ext",
    [
        ("audio", "wav", "textgrid", ".wav"),
    ],
)
def test__map_replacing_ext(
    source_type,
    source_format,
    annot_format,
    annotated_ext,
    audio_list_factory,
    spect_list_mat,
    specific_annot_list,
):
    if source_type == "audio":
        annotated_files = audio_list_factory(source_format, annot_format)
    else:
        annotated_files = spect_list_mat
    annot_list = specific_annot_list(annot_format)

    annotated_annot_map = vak.annotation._map_replacing_ext(
        annotated_files=annotated_files, annot_list=annot_list, annotated_ext=annotated_ext
    )

    # test all the audio paths made it into the map
    annotated_files_from_map = list(annotated_annot_map.keys())
    for source_file in annotated_files:
        assert source_file in annotated_files_from_map

    # test all the annots made it into the map
    annot_list_from_map = list(annotated_annot_map.values())
    for annot in annot_list:
        assert annot in annot_list_from_map

    # test all mappings are correct
    for source_path, annot in list(annotated_annot_map.items()):
        assert vak.annotation.audio_stem_from_path(
            annot.audio_path
        ) == vak.annotation.audio_stem_from_path(source_path)


@pytest.mark.parametrize(
    "source_type, source_format, annot_format",
    [
        ("audio", "cbin", "notmat"),
        ("audio", "cbin", "simple-seq"),
        ("audio", "wav", "birdsong-recognition-dataset"),
        ("audio", "wav", "textgrid"),
        ("spect", "mat", "yarden"),
        ("audio", "wav", "textgrid"),
    ],
)
def test_map_annotated_to_annot(
    source_type,
    source_format,
    annot_format,
    audio_list_factory,
    spect_list_mat,
    specific_annot_list,
):
    if source_type == "audio":
        annotated_files = audio_list_factory(source_format, annot_format)
    else:
        annotated_files = spect_list_mat
    annot_list = specific_annot_list(annot_format)
    annotated_annot_map = vak.annotation.map_annotated_to_annot(
        annotated_files=annotated_files, annot_list=annot_list
    )

    # test all the audio paths made it into the map
    annotated_files_from_map = list(annotated_annot_map.keys())
    for source_file in annotated_files:
        assert source_file in annotated_files_from_map

    # test all the annots made it into the map
    annot_list_from_map = list(annotated_annot_map.values())
    for annot in annot_list:
        assert annot in annot_list_from_map

    # test all mappings are correct
    for source_path, annot in list(annotated_annot_map.items()):
        assert vak.annotation.audio_stem_from_path(
            annot.audio_path
        ) == vak.annotation.audio_stem_from_path(source_path)


@pytest.mark.parametrize(
    'annot, duration, expected_result',
    [
        # common case: unlabeled periods between annotated segments, and before and after.
        # Function should return true.
        (
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=np.array([1.0, 2.0, 3.0, 4.0]),
                    offsets_s=np.array([1.5, 2.5, 3.5, 4.5]),
                    labels=np.array(['a', 'b', 'c', 'd'])),
                    annot_path='/dummy/annot/path/annot.csv'),
            5.0, True
        ),
        # other common case we expect: all periods are labeled. Function should return False.
        (
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
                    offsets_s=np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
                    labels=np.array(['a', 'b', 'c', 'd'])),
                    annot_path='/dummy/annot/path/annot.csv'),
                5.0, False
        ),
        # edge case 1: only unlabeled period is before annotated segments
        (
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=np.array([1.0, 2.0, 3.0, 4.0]),
                    offsets_s=np.array([2.0, 3.0, 4.0, 5.0]),
                    labels=np.array(['a', 'b', 'c', 'd'])),
                annot_path='/dummy/annot/path/annot.csv'),
                5.0, True
        ),
        # edge case 2: only unlabeled period is after annotated segments
        (
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
                    offsets_s=np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
                    labels=np.array(['a', 'b', 'c', 'd'])),
                annot_path='/dummy/annot/path/annot.csv'),
                6.0, True
        ),
    ]
)
def test_has_unlabeled(annot, duration, expected_result):
    """Test ``vak.annotation.has_unlabeled``,
    including edge cases as discussed in https://github.com/vocalpy/vak/issues/243
    """
    has_unlabeled = vak.annotation.has_unlabeled(annot, duration)
    assert has_unlabeled == expected_result


def test_has_unlabeled_annotation_with_no_segments(annotated_annot_no_segments):
    """Test edge case for `has_unlabeled`,
    see https://github.com/vocalpy/vak/issues/378
    """
    audio_path, annot_path = annotated_annot_no_segments
    data, samp_freq = evfuncs.load_cbin(audio_path)
    dur = data.shape[0] / samp_freq
    scribe = crowsetta.Transcriber(format='notmat')
    annot = scribe.from_file(annot_path)

    assert vak.annotation.has_unlabeled(annot, dur) is True



"""fixtures relating to annotation files"""
import crowsetta
import pytest
import toml


from .test_data import SOURCE_TEST_DATA_ROOT


@pytest.fixture
def annot_file_yarden(source_test_data_root):
    return source_test_data_root.joinpath(
        "spect_mat_annot_yarden", "llb3", "llb3_annot_subset.mat"
    )


@pytest.fixture
def annot_list_yarden(annot_file_yarden):
    scribe = crowsetta.Transcriber(format="yarden")
    annot_list = scribe.from_file(annot_file_yarden)
    return annot_list


@pytest.fixture
def labelset_yarden():
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set
    """
    return [
        str(an_int)
        for an_int in [1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 18, 19]
    ]


@pytest.fixture
def annot_dir_notmat(source_test_data_root):
    return source_test_data_root.joinpath("audio_cbin_annot_notmat", "gy6or6", "032312")


@pytest.fixture
def annot_files_notmat(annot_dir_notmat):
    return sorted(annot_dir_notmat.glob("*.not.mat"))


@pytest.fixture
def annot_list_notmat(annot_files_notmat):
    scribe = crowsetta.Transcriber(format="notmat")
    annot_list = scribe.from_file(annot_files_notmat)
    return annot_list


@pytest.fixture
def labelset_notmat(generated_test_configs_root):
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set"""
    a_train_notmat_config = sorted(
        generated_test_configs_root.glob("*train*notmat*toml")
    )[
        0
    ]  # get first config.toml from glob list
    # doesn't really matter which config, they all have labelset
    with a_train_notmat_config.open("r") as fp:
        a_train_notmat_toml = toml.load(fp)
    labelset = a_train_notmat_toml["PREP"]["labelset"]
    return labelset


@pytest.fixture
def annot_file_birdsongrec(source_test_data_root):
    return source_test_data_root.joinpath(
        "audio_wav_annot_birdsongrec", "Bird0", "Annotation.xml"
    )


@pytest.fixture
def annot_list_birdsongrec(annot_file_birdsongrec):
    scribe = crowsetta.Transcriber(format="birdsong-recognition-dataset")
    annot_list = scribe.from_file(annot_file_birdsongrec)
    return annot_list


@pytest.fixture
def annot_dir_textgrid(source_test_data_root):
    return source_test_data_root.joinpath("audio_wav_annot_textgrid", "AGBk")


@pytest.fixture
def annot_files_textgrid(annot_dir_textgrid):
    return sorted(annot_dir_textgrid.glob("*.TextGrid"))


@pytest.fixture
def annot_list_textgrid(annot_files_textgrid):
    scribe = crowsetta.Transcriber(format="textgrid")
    annot_list = scribe.from_file(annot_files_textgrid)
    return annot_list


@pytest.fixture
def annot_dir_simple_seq(source_test_data_root):
    return source_test_data_root.joinpath("audio_cbin_annot_simple_seq", "gy6or6", "032312")


@pytest.fixture
def annot_files_simple_seq(annot_dir_simple_seq):
    return sorted(annot_dir_simple_seq.glob("*.cbin.csv"))


@pytest.fixture
def annot_list_simple_seq(annot_files_simple_seq):
    scribe = crowsetta.Transcriber(format="simple-seq")
    annot_list = scribe.from_file(annot_files_simple_seq)
    return annot_list


@pytest.fixture
def specific_annot_list(annot_list_birdsongrec,
                        annot_list_notmat,
                        annot_list_simple_seq,
                        annot_list_textgrid,
                        annot_list_yarden):
    """factory fixture, returns a function that
    returns a fixture containing a list of Annotation objects,
    given a specified annotation format

    so that unit tests can be parameterized with annotation format names
    """
    FORMAT_ANNOT_LIST_FIXTURE_MAP = {
        "birdsong-recognition-dataset": annot_list_birdsongrec,
        "notmat": annot_list_notmat,
        "simple-seq": annot_list_simple_seq,
        "textgrid": annot_list_textgrid,
        "yarden": annot_list_yarden,
    }

    def _annot_list_factory(format):
        return FORMAT_ANNOT_LIST_FIXTURE_MAP[format]

    return _annot_list_factory


@pytest.fixture
def specific_labelset(labelset_yarden, labelset_notmat):
    def _specific_labelset(annot_format):
        if annot_format == "yarden":
            return labelset_yarden
        elif annot_format == "notmat":
            return labelset_notmat
        else:
            raise ValueError(f"invalid annot_format: {annot_format}")

    return _specific_labelset


ANNOT_FILES_WITH_NO_SEGMENTS_DIR = SOURCE_TEST_DATA_ROOT / 'audio_cbin_annot_notmat' / 'gy6or6-annotated-with-no-labels'
ANNOTATED_FILES_NO_SEGMENTS = sorted(ANNOT_FILES_WITH_NO_SEGMENTS_DIR.glob('*.cbin'))
ANNOT_FILES_WITH_NO_SEGMENTS = sorted(ANNOT_FILES_WITH_NO_SEGMENTS_DIR.glob('*.not.mat'))
ANNOTATED_ANNOT_NO_SEGMENTS_TUPLES = list(
    zip(ANNOTATED_FILES_NO_SEGMENTS, ANNOT_FILES_WITH_NO_SEGMENTS)
)


@pytest.fixture(params=ANNOTATED_ANNOT_NO_SEGMENTS_TUPLES)
def annotated_annot_no_segments(request):
    """Tuple of (annotated (audio) path, annotation path),
    where the annotation file has no annotated segments in it.
    Used to test edge case for `has_unlabeled`,
    see https://github.com/vocalpy/vak/issues/378
    """
    return request.param



"""tests for vak.files.spect module"""
import itertools
import pathlib

import pytest

import vak.files
from vak.constants import VALID_AUDIO_FORMATS


from ..fixtures.spect import (
    SPECT_LIST_MAT,
    SPECT_LIST_NPZ,
)


SPECT_LIST_MAT = SPECT_LIST_MAT + [
    # duplicate list but replace underscores in filenames with spaces to test handling spaces
    spect_path.parent / spect_path.name.replace('_', ' ')
    for spect_path in SPECT_LIST_MAT
] + [
    # duplicate list but replace underscores in entire path with spaces to test handling spaces
    pathlib.Path(str(spect_path).replace('_', ' '))
    for spect_path in SPECT_LIST_MAT
]
# test function with string paths not just pathlib.Path instances
SPECT_LIST_MAT = SPECT_LIST_MAT + [str(path) for path in SPECT_LIST_MAT]

SPECT_LIST_NPZ = SPECT_LIST_NPZ + [
    # duplicate list but replace underscores in filenames with spaces to test handling spaces
    spect_path.parent / spect_path.name.replace('_', ' ')
    for spect_path in SPECT_LIST_NPZ
] + [
    # duplicate list but replace underscores in entire path with spaces to test handling spaces
    pathlib.Path(str(spect_path).replace('_', ' '))
    for spect_path in SPECT_LIST_NPZ
]
# test function with string paths not just pathlib.Path instances
SPECT_LIST_NPZ = SPECT_LIST_NPZ + [str(path) for path in SPECT_LIST_NPZ]

SPECT_LIST_MAT_WITH_EXT = list(zip(
    SPECT_LIST_MAT,
    itertools.repeat('.wav'),
    itertools.repeat('.mat'),
)) + list(zip(
    SPECT_LIST_MAT,
    # also test with case where we don't specify audio extension
    itertools.repeat(None),
    itertools.repeat('.mat'),
))

SPECT_LIST_NPZ_WITH_EXT = list(zip(
    SPECT_LIST_NPZ,
    itertools.repeat('.cbin'),
    itertools.repeat('.spect.npz'),
)) + list(zip(
    SPECT_LIST_NPZ,
    itertools.repeat(None),
    itertools.repeat('.spect.npz'),
))

TEST_FIND_AUDIO_FNAME_PARAMETRIZE = SPECT_LIST_MAT_WITH_EXT + SPECT_LIST_NPZ_WITH_EXT


@pytest.mark.parametrize(
    "spect_path, audio_ext, spect_ext",
    TEST_FIND_AUDIO_FNAME_PARAMETRIZE
)
def test_find_audio_fname_with_mat(spect_path, audio_ext, spect_ext):
    """test ```vak.files.spect.find_audio_fname`` works when we give it a list of """
    expected = pathlib.Path(spect_path).name.replace(spect_ext, '')

    out = vak.files.spect.find_audio_fname(spect_path, audio_ext)

    # make sure we gout out a filename that was actually in spect_path
    assert pathlib.Path(spect_path).name.startswith(out)
    # make sure it's some valid audio format
    assert pathlib.Path(out).suffix.replace(".", "") in VALID_AUDIO_FORMATS
    assert out == expected



import numpy as np
import pytest

import vak.transforms
import vak.transforms.functional
import vak.validators


class TestStandardizeSpect:

    @pytest.mark.parametrize(
        'mean_freqs, std_freqs, non_zero_std',
        [
            (
                np.array([0., 0., 0., 0., 0.]),
                np.array([1., 1., 1., 1., 1.]),
                np.array([0, 1, 2, 3, 4]),
            )
        ]
    )
    def test_instance(self, mean_freqs, std_freqs, non_zero_std):
        standardizer = vak.transforms.StandardizeSpect(
            mean_freqs=mean_freqs, std_freqs=std_freqs, non_zero_std=non_zero_std
        )
        assert isinstance(standardizer, vak.transforms.StandardizeSpect)
        for attr_name, expected in zip(
                ('mean_freqs', 'std_freqs', 'non_zero_std',),
                (mean_freqs, std_freqs, non_zero_std),
        ):
            assert hasattr(standardizer, attr_name)
            attr = getattr(standardizer, attr_name)
            assert isinstance(attr, np.ndarray)
            assert np.all(
                np.equal(attr, expected)
            )

        T_BINS = 100
        spect = np.random.rand(mean_freqs.shape[0], T_BINS)
        spect_out = standardizer(spect.copy())
        expected = vak.transforms.functional.standardize_spect(spect, mean_freqs, std_freqs, non_zero_std)
        assert np.all(
            np.equal(spect_out, expected)
        )

    @pytest.mark.parametrize(
        'split',
        [
            'train',
            'val',
            None
        ]
    )
    def test_fit_df(self, split, train_cbin_notmat_df):
        if split is None:
            split_to_test = 'train'
        else:
            split_to_test = split
        # ---- set up
        df_split = train_cbin_notmat_df[train_cbin_notmat_df.split == split_to_test].copy()
        spect_paths = df_split['spect_path'].values
        spect = vak.files.spect.load(spect_paths[0])['s']
        mean_freqs = np.mean(spect, axis=1)
        std_freqs = np.std(spect, axis=1)

        for spect_path in spect_paths[1:]:
            spect = vak.files.spect.load(spect_path)['s']
            mean_freqs += np.mean(spect, axis=1)
            std_freqs += np.std(spect, axis=1)
        expected_mean_freqs = mean_freqs / len(spect_paths)
        expected_std_freqs = std_freqs / len(spect_paths)
        expected_non_zero_std = np.argwhere(expected_std_freqs != 0)
        # we convert to 1d vector in __init__
        expected_non_zero_std = vak.validators.column_or_1d(expected_non_zero_std)

        # ---- actually do fit
        if split:
            standardizer = vak.transforms.StandardizeSpect.fit_df(train_cbin_notmat_df,
                                                                  split=split)
        else:
            # this tests that default value for split 'train' works as expected
            standardizer = vak.transforms.StandardizeSpect.fit_df(train_cbin_notmat_df)

        # ---- test
        for attr_name, expected in zip(
                ('mean_freqs', 'std_freqs', 'non_zero_std',),
                (expected_mean_freqs, expected_std_freqs, expected_non_zero_std),
        ):
            assert hasattr(standardizer, attr_name)
            attr = getattr(standardizer, attr_name)
            assert isinstance(attr, np.ndarray)
            assert np.all(
                np.equal(attr, expected)
            )

    def test_fit(self):
        spect = np.random.rand(513, 1000)
        standardizer = vak.transforms.StandardizeSpect.fit(spect)

        expected_mean_freqs = np.mean(spect, axis=1)
        expected_std_freqs = np.std(spect, axis=1)
        expected_non_zero_std = np.argwhere(expected_std_freqs != 0)
        # we convert to 1d vector in __init__
        expected_non_zero_std = vak.validators.column_or_1d(expected_non_zero_std)

        for attr_name, expected in zip(
                ('mean_freqs', 'std_freqs', 'non_zero_std',),
                (expected_mean_freqs, expected_std_freqs, expected_non_zero_std),
        ):
            assert hasattr(standardizer, attr_name)
            attr = getattr(standardizer, attr_name)
            assert isinstance(attr, np.ndarray)
            assert np.all(
                np.equal(attr, expected)
            )



"""tests for vak.config.eval module"""
import vak.config.eval


def test_predict_attrs_class(all_generated_eval_configs_toml):
    """test that instantiating EvalConfig class works as expected"""
    for config_toml in all_generated_eval_configs_toml:
        eval_section = config_toml["EVAL"]
        config = vak.config.eval.EvalConfig(**eval_section)
        assert isinstance(config, vak.config.eval.EvalConfig)



def eval(toml_path):
    from .eval import eval
    eval(toml_path=toml_path)


def train(toml_path):
    from .train import train
    train(toml_path=toml_path)


def learncurve(toml_path):
    from .learncurve import learning_curve
    learning_curve(toml_path=toml_path)


def predict(toml_path):
    from .predict import predict
    predict(toml_path=toml_path)


def prep(toml_path):
    from .prep import prep
    prep(toml_path=toml_path)


COMMAND_FUNCTION_MAP = {
    "prep": prep,
    "train": train,
    "eval": eval,
    "predict": predict,
    "learncurve": learncurve,
}


CLI_COMMANDS = tuple(COMMAND_FUNCTION_MAP.keys())


def cli(command, config_file):
    """Execute the commands of the command-line interface.

    Parameters
    ----------
    command : string
        One of {'prep', 'train', 'eval', 'predict', 'learncurve'}
    config_file : str, Path
        path to a config.toml file
    """
    if command in COMMAND_FUNCTION_MAP:
        COMMAND_FUNCTION_MAP[command](toml_path=config_file)
    else:
        raise ValueError(f"command not recognized: {command}")



"""Validators for sequence datasets"""
import numpy as np
import pandas as pd

from ... import annotation, files


def where_unlabeled(csv_path, timebins_key="t"):
    """Find where there are vocalizations in a dataset
    annotated as sequences that have unlabeled periods
    in those annotations.

    Parameters
    ----------
    csv_path : str, Path
        Path to .csv file representing dataset.
    timebins_key : str
        Key used to access timebins vector in spectrogram files.
        Default is 't'.

    Returns
    -------
    where_unlabeled : numpy.ndarray
        Vector with Boolean dtype, where a True
        element indicates that the
        vocalization indexed by this has
        periods that are unlabeled by the annotations.
    """
    vak_df = pd.read_csv(csv_path)
    annots = annotation.from_df(vak_df)

    has_unlabeled_list = []
    for annot, spect_path in zip(annots, vak_df["spect_path"].values):
        spect_dict = files.spect.load(spect_path)
        timebins = spect_dict[timebins_key]
        duration = timebins[-1]

        has_unlabeled_list.append(
            annotation.has_unlabeled(annot, duration)
        )

    return np.array(has_unlabeled_list)


def has_unlabeled(csv_path, timebins_key="t"):
    r"""Determine if a dataset of vocalization
    annotated as sequences has periods that are unlabeled.

    Used to decide whether an additional class needs to be added
    to the set of labels :math:`Y = {y_1, y_2, \dots, y_n}`,
    where the added class :math:`y_{n+1}`
    will represent the unlabeled "background" periods.

    Parameters
    ----------
    csv_path : str, Path
        Path to .csv file representing dataset.
    timebins_key : str
        Key used to access timebins vector in spectrogram files.
        Default is 't'.

    Returns
    -------
    has_unlabeled : bool
        If True, there are annotations in the dataset
        that have unlabeled periods.
    """
    return np.any(
        where_unlabeled(csv_path, timebins_key)
    )



from __future__ import annotations
from collections import Counter
import copy
import os
from pathlib import Path
from typing import Optional, Union

import crowsetta
import numpy as np

from . import files
from . import constants
from .typing import PathLike


def format_from_df(vak_df):
    """determine annotation format of a Vak DataFrame.
    Returns string name of annotation format.
    If no annotation format is specified, returns None.
    Raises an error if there are multiple formats.

    Parameters
    ----------
    vak_df : pandas.DataFrame
        Representing a dataset of vocalizations,
        with column 'annot_format'.

    Returns
    -------
    annot_format : str
        format of annotations for vocalizations.
    """
    annot_format = vak_df["annot_format"].unique()
    if len(annot_format) == 1:
        annot_format = annot_format.item()
        if annot_format is None or annot_format == constants.NO_ANNOTATION_FORMAT:
            return None
    elif len(annot_format) > 1:
        raise ValueError(
            f"unable to load labels for dataset, found multiple annotation formats: {annot_format}"
        )

    return annot_format


def from_df(vak_df):
    """get list of annotations from a vak DataFrame.
    If no annotation format is specified for the DataFrame
    (in the 'annot_format' column), returns None.

    Parameters
    ----------
    vak_df : DataFrame
        representating a dataset of vocalizations, with column 'annot_format'.

    Returns
    -------
    annots : list
        of annotations for each row in the dataframe,
        represented as crowsetta.Annotation instances.

    Notes
    -----
    This function encapsulates logic for handling different types of
    annotations; it determines whether each row has a separate annotation file,
    or if instead there is a single annotation file associated with all rows.
    If the latter, then the function opens that file and makes sure that
    each row from the dataframe can be paired with an annotation (using `map_annotated_to_annot`).
    """
    annot_format = format_from_df(vak_df)
    if annot_format is None:
        return None

    scribe = crowsetta.Transcriber(format=annot_format)

    if len(vak_df["annot_path"].unique()) == 1:
        # --> there is a single annotation file associated with all rows
        # this can be true in two different cases:
        # (1) many rows, all have the same file
        # (2) only one row, so there's only one annotation file (which may contain annotation for multiple source files)
        annot_path = vak_df["annot_path"].unique().item()
        annots = scribe.from_file(annot_path)

        # as long as we have at least as many annotations as there are rows in the dataframe
        if (isinstance(annots, list) and len(annots) >= len(vak_df)) or (  # case 1
            isinstance(annots, crowsetta.Annotation) and len(vak_df) == 1
        ):  # case 2
            if isinstance(annots, crowsetta.Annotation):
                annots = [
                    annots
                ]  # wrap in list for map_annotated_to_annot to iterate over it
            # then we can try and map those annotations to the rows
            audio_annot_map = map_annotated_to_annot(vak_df["audio_path"].values, annots)
            # sort by row of dataframe
            annots = [
                audio_annot_map[audio_path]
                for audio_path in vak_df["audio_path"].values
            ]

        else:
            raise ValueError(
                "unable to load labels from dataframe; found a single annotation file associated with all "
                "rows in dataframe, but loading it did not return a list of annotations for each row.\n"
                f"Single annotation file: {annot_path}\n"
                f"Loading it returned a {type(annots)}."
            )

    elif len(vak_df["annot_path"].unique()) == len(vak_df):
        # --> there is a unique annotation file (path) for each row, iterate over them to get labels from each
        annots = [
            scribe.from_file(annot_path) for annot_path in vak_df["annot_path"].values
        ]

    else:
        raise ValueError(
            "unable to load labels from dataframe; did not find an annotation file for each row or "
            "a single annotation file associated with all rows."
        )

    return annots


def files_from_dir(annot_dir, annot_format):
    """get all annotation files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.
    """
    if annot_format not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f"specified annotation format, {annot_format} not valid.\n"
            f"Valid formats are: {constants.VALID_ANNOT_FORMATS}"
        )

    format_module = getattr(crowsetta.formats, annot_format)
    ext = format_module.meta.ext
    annot_files = files.from_dir(annot_dir, ext)
    return annot_files


class AudioFilenameNotFound(Exception):
    """Error raised when a name of an audio filename
    cannot be found within another filename.

    Raised by ``audio_stem_from_path``
    and ``_map_using_audio_stem_from_path``.
    """


def audio_stem_from_path(path: PathLike,
                         audio_ext: str = None) -> str:
    """Find the name of an audio file within a filename
    by removing extensions until finding an audio extension,
    then return the name of that audio file
    without the extension (i.e., the "stem").

    Removes extensions from a filename recursively,
    by calling `os.path.splitext`,
    until the extension is an audio file format handled by vak.
    Then return the stem, that is,
    the part that precedes the extension.
    Used to match audio, spectrogram,
    and annotation files by their stems.

    Stops after finding audio extensions
    so that it does not remove "extensions"
    that are actually other parts of a filename,
    e.g. a time or data separated by periods.

    Examples
    --------
    >>> audio_stem_from_path('gy6or6_baseline_230312_0808.138.cbin.not.mat')
    'gy6or6_baseline_230312_0808.138'
    >>> audio_stem_from_path('Bird0/spectrograms/0.wav.npz')
    '0'
    >>> audio_stem_from_path('Bird0/Wave/0.wav')
    '0'

    Parameters
    ----------
    path : str, Path
        Path to a file that contains an audio filename in its name.
    audio_ext : str
        Extension corresponding to format of audio file.
        Must be one of ``vak.contsants.VALID_AUDIO_FORMATS``.
        Default is None, in which case the function looks
        removes extensions until it finds any valid audio
        format extension.

    Returns
    -------
    stem : str
        Part of filename that precedes audio extension.
    """
    if audio_ext:
        if audio_ext.startswith('.'):
            audio_ext = audio_ext[1:]
        if audio_ext not in constants.VALID_AUDIO_FORMATS:
            raise ValueError(
                f'Not a valid extension for audio formats: {audio_ext}\n'
                f'Valid formats are: {constants.VALID_AUDIO_FORMATS}'
            )
        extensions_to_look_for = [audio_ext]
    else:
        extensions_to_look_for = constants.VALID_AUDIO_FORMATS

    name = Path(path).name
    stem, ext = os.path.splitext(name)
    ext = ext.replace(".", "").lower()
    while ext not in extensions_to_look_for:
        new_stem, ext = os.path.splitext(stem)
        ext = ext.replace(".", "").lower()
        if new_stem == stem:
            raise AudioFilenameNotFound(
                f"Unable to find a valid audio filename in path:\n{path}.\n"
                f"Valid audio file extensions are:\n{constants.VALID_AUDIO_FORMATS}"
            )
        else:
            stem = new_stem
    return stem


def _map_using_audio_stem_from_path(annotated_files: list[PathLike],
                                    annot_list: list[crowsetta.Annotation],
                                    audio_ext: Optional[str] = None) -> dict:
    """Map a list of annotated files to a list of ``crowsetta.Annotation``s,
    using the ``audio_path`` attribute of the ``Annotation``s.

    One of two helper functions used by ``map_annotated_to_annot``.

    This function assumes each file in ``annotated_files``
    contains an audio path filename in it, so that
    the ``audio_stem_from_path`` function can be used to find
    the stem, and match it with the stem of the
    ``audio_path`` attribute of an ``Annotation`` instance.
    In other words, we map by matching the following:
    stem_from_audio_file(annotation.audio_path) <--> stem_from_audio_file(annotated_file)

    Parameters
    ----------
    annotated_files : list
        List of paths to the annotated files.
    annot_list : list
        List of ``crowsetta.Annotation`` instances.
    audio_ext : str
        Extension corresponding to audio format.
        Valid extension are listed in
        ``vak.constants.VALID_AUDIO_FORMATS``.
        Default is None, in which case the function
        looks for any valid format.

    Returns
    -------
    annotated_annot_map : dict
        Where each key is path to annotated file, and
        its value is the corresponding ``crowsetta.Annotation``.
    """
    # First check that we don't have duplicate keys that would cause this to fail silently
    keys = []
    for annot in annot_list:
        try:
            stem = audio_stem_from_path(annot.audio_path, audio_ext)
        except AudioFilenameNotFound as e:
            # Do this as a loop with a super verbose error
            # instead of e.g. a single-line list comprehension
            # so we can help users troubleshoot,
            # see https://github.com/vocalpy/vak/issues/525
            raise AudioFilenameNotFound(
                "The ``audio_path`` attribute of a ``crowsetta.Annotation`` was "
                "not recognized as a valid audio filename.\n"
                f"The ``audio_path`` attribute was:\n{annot.audio_path}\n"
                f"The annotation was loaded from this path:\n{annot.annot_path}\n"
            ) from e
        keys.append(stem)

    keys_set = set(keys)
    if len(keys_set) < len(keys):
        duplicates = [item for item, count in Counter(keys).items() if count > 1]
        raise ValueError(
            f"found multiple annotations with the same audio filename(s): {duplicates}"
        )
    del keys, keys_set
    # ----> make a dict with audio stems as keys,
    #       so we can look up annotations
    #       by stemming source files and using as keys.
    audio_stem_annot_map = {
        # NOTE HERE WE GET STEMS FROM EACH annot.audio_path,
        # BELOW we get stems from each annotated_file
        audio_stem_from_path(annot.audio_path): annot for annot in annot_list
    }

    # Make a copy of ``annotated_files`` from which
    # we remove files after mapping them to annotation,
    # to validate that function worked,
    # by making sure there are no items left in this copy after the loop.
    # If there is 1:1 mapping then there should be no items left.
    annotated_annot_map = {}
    annotated_files_copy = copy.deepcopy(annotated_files)
    for annotated_file in annotated_files:
        # stem annotated file so we can find audio OR spect files
        # that match with stems from each annot.audio_path;
        # e.g. find '~/path/to/llb3/llb3_0003_2018_04_23_14_18_54.wav.mat' that
        # should match with ``Annotation(audio_path='llb3_0003_2018_04_23_14_18_54.wav')``
        annotated_file_stem = audio_stem_from_path(annotated_file)
        annot = audio_stem_annot_map[annotated_file_stem]
        annotated_annot_map[annotated_file] = annot
        annotated_files_copy.remove(annotated_file)

    if len(annotated_files_copy) > 0:
        raise ValueError(
            "Could not map the following source files to annotations: "
            f"{annotated_files_copy}"
        )
    return annotated_annot_map


class AnnotatedFilenameNotFound(Exception):
    """Exception raised when a name for
    an annotated file cannot be matched
    with its annotation,
    by replacing the extension of
    the annotation file with the
    extension of the annotated
    file format.

    Raised by ``_map_replacing_ext``.
    """
    pass


def _map_replacing_ext(annotated_files: list[PathLike],
                       annot_list: list[crowsetta.Annotation],
                       annotated_ext: str) -> dict[crowsetta.Annotation: PathLike]:
    """Map a list of annotated files
    to a list of ``crowsetta.Annotation``s
    using an extension
    for the annotated files, ``annotated_ext``.

    One of two helper functions used by ``map_annotated_to_annot``.

    First, replaces the extension of each
    ``Annotation``s ``annot_path`` attribute
    with ``annotated_ext``, and then
    verifies that each path generated
    this way can be matched with a path in
    ``annotated_files``, without any
    missing or any extras.

    Parameters
    ----------
    annotated_files : list
        List of paths to the annotated files.
    annot_list : list
        List of ``crowsetta.Annotation`` instances.
    annotated_ext : str
        Extension of the annotated files.
        E.g., an audio format like 'wav',
        or an array file format like 'mat'
        or 'npz' that can contain spectrograms.

    Returns
    -------
    annotated_annot_map : dict
        Where each key is path to annotated file, and
        its value is the corresponding ``crowsetta.Annotation``.
    """
    # sanitize arg, so we are always working with
    # lowercase extension w/out period
    annotated_ext = annotated_ext.lower()
    if annotated_ext.startswith('.'):
        annotated_ext = annotated_ext[1:]

    # we make two lists that we will match:
    # one from annotations, the other from annotated files
    annot_filenames_with_annotated_ext = [
        annot.annot_path.stem + f'.{annotated_ext}'
        for annot in annot_list
    ]
    annotated_files_just_names = [
        # make actual annotated file extension lower too; ensures we can match
        Path(annotated_file).stem + Path(annotated_file).suffix.lower()
        for annotated_file in annotated_files
    ]
    # Make a copy of annotated files
    # from which we remove entries after mapping them to an annotation,
    # to validate that function worked,
    # by making sure there are no items left in this copy after the loop.
    # If there is 1:1 mapping then there should be no items left.
    annotated_files_copy = copy.deepcopy(annotated_files)
    annotated_annot_map = {}
    for annot, annot_filename_with_annotated_ext in zip(annot_list,
                                                    annot_filenames_with_annotated_ext):
        try:
            idx = annotated_files_just_names.index(
                annot_filename_with_annotated_ext
            )
        except ValueError as e:
            raise AnnotatedFilenameNotFound(
                f"Unable to match annotation with annotated file, "
                f"using the annotation file's path:\n{annot.annot_path}\n"
                f"The expected name of the annotated file, "
                f"using the ``annotated_ext`` '{annotated_ext}`, "
                f"was:\n{annot_filename_with_annotated_ext}"
            ) from e
        annotated_annot_map[annotated_files[idx]] = annot
        annotated_files_copy.remove(annotated_files[idx])

    if len(annotated_files_copy) > 0:
        raise ValueError(
            "Could not map the following source files to annotations: "
            f"{annotated_files_copy}"
        )
    return annotated_annot_map


def map_annotated_to_annot(annotated_files: Union[list, np.array],
                           annot_list: list[crowsetta.Annotation],
                           audio_ext: Optional[str] = None,
                           annotated_ext: Optional[str] = None) -> dict:
    """Map annotated files,
    i.e. audio or spectrogram files,
    to their corresponding annotations.

    Returns a ``dict`` where each key
    is a path to an annotated file,
    and the value for each key
    is a ``crowsetta.Annotation``.

    Mapping is done with two helper functions,
    ``_map_using_audio_stem_from_path`` and
    ``_map_replacing_ext``. The first assumes
    that the annotated function contains
    the name of the original audio file,
    and that this can be matched with the
    ``audio_path`` attribute of one of the
    ``Annotation`` instances. The second,
    which is used if the first fails,
    tries replacing the extension of the
    annotation files with a specified
    extension of the annotated files.

    Parameters
    ----------
    annotated_files : list
        Of paths to audio or spectrogram files.
    annot_list : list
        Of Annotations corresponding to files in annotated_files
    audio_ext : str
        Extension of audio files.
        Default is None, in which case this function will
        look for extensions of any valid audio format
        (listed as ``vak.constants.VALID_AUDIO_FORMAT``).
        Specifying the format provides a slight speed up.
    annotated_ext : str
        Extension of the annotated files.
        Default is None
    Notes
    -----
    The filenames of the ``annotated_files`` must
    begin with the filename of the ``audio_path``
    attribute of the corresponding
    ``crowsetta.Annotation`` instances.
    E.g., if `annotated_files` includes
    an audio file named
    'bird0-2016-05-04-133027.wav',
    then it will be mapped to an ``Annotation``
    with an `audio_path` attribute
    whose filename matches it.
    Spectrogram files should also include
    the audio file name,
    e.g. 'bird0-2016-05-04-133027.wav.mat'
    or 'bird0-2016-05-04-133027.spect.npz'
    would match an ``Annotation`` with the
    ``audio_path`` attribute '/some/path/bird0-2016-05-04-133027.wav'.

    For more detail, please see
    the page on file naming conventions in the
    reference section of the documentation:
    https://vak.readthedocs.io/en/latest/reference/filenames.html
    """
    if type(annotated_files) == np.ndarray:  # e.g., vak DataFrame['spect_path'].values
        annotated_files = annotated_files.tolist()

    try:
        annotated_annot_map = _map_using_audio_stem_from_path(annotated_files,
                                                              annot_list,
                                                              audio_ext)
    except AudioFilenameNotFound:
        try:
            annotated_annot_map = _map_replacing_ext(annotated_files,
                                                     annot_list,
                                                     annotated_ext)
        except AnnotatedFilenameNotFound as e:
            raise ValueError(
                'Could not map annotated files to annotations.\n'
                'Please see this section in the `vak` documentation:\n'
                'https://vak.readthedocs.io/en/latest/howto/howto_prep_annotate.html#how-does-vak-know-which-annotations-go-with-which-annotated-files'
            ) from e

    return annotated_annot_map


def has_unlabeled(annot: crowsetta.Annotation,
                  duration: float):
    """Returns ``True`` if an annotated sequence has unlabeled segments.

    Tests whether an instance of ``crowsetta.Annotation.seq`` has
    intervals between the annotated segments with a non-zero duration,
    or any unannotated periods before or after the annotated segments.

    Parameters
    ----------
    annot : crowsetta.Annotation
        With a ``seq`` attribute that is a ``crowsetta.Sequence``
    duration : float
        Total duration of the vocalization
        that is annotated by ``annot``.
        Needed to determine whether the duration
        is greater than the time
        of the last offset in the annotated segments.

    Returns
    -------
    has_unlabeled : bool
        If True, there are unlabeled periods
        in the vocalization annotated by ``annot``.
    """
    if duration <= 0:
        raise ValueError(
            f"Duration less than or equal to zero passed to `has_unlabeled`.\n"
            f"Value for `duration`: {duration}.\nValue for `annot`: {annot}"
        )
    if duration > 0 and len(annot.seq.segments) < 1:
        # Handle edge case where there are no annotated segments in annotation file
        # See https://github.com/vocalpy/vak/issues/378
        return True
    has_unlabeled_intervals = np.any((annot.seq.onsets_s[1:] - annot.seq.offsets_s[:-1]) > 0.)
    has_unlabeled_before_first_onset = annot.seq.onsets_s[0] > 0.
    has_unlabeled_after_last_offset = duration - annot.seq.offsets_s[-1] > 0.
    return has_unlabeled_intervals or has_unlabeled_before_first_onset or has_unlabeled_after_last_offset



"""utility functions for logging"""
import logging
from pathlib import Path
import sys
import warnings

from . import timenow
from .__about__ import __version__


logger = logging.getLogger('vak')  # 'base' logger


def config_logging_for_cli(log_dst: str,
                           log_stem: str,
                           level='info',
                           timestamp=None,
                           force=False):
    """Configure logging for a run of the cli.

    Called by `vak.cli` functions. Allows logging
    to also be configured differently by a user
    that interacts directly with the library through e.g. `vak.core`.

    Parameters
    ----------
    log_dst : str, Path
        Destination directory where log file should be saved
    log_stem : str
        The "stem" of the filename for the log file.
        A timestamp is added to this stem,
        so that the final filename is ``f"{caller}_{timestamp}.log"``.
        Usually set to the name of the function that called this one,
        e.g. 'train' or 'prep'.
    level : str
        Logging level, e.g. "INFO". Passed in to ``logger.setLevel``.
    timestamp : str
        Time stamp, included in log file name.
        If None, defaults to `datetime.now().strftime('%y%m%d_%H%M%S')`.
    force : bool
        If True, forces this function to remove the old handlers
        and add new handlers. If False, this function will raise an
        error when the logger already has handlers,
        to avoid silent failures.
    """
    log_dst = Path(log_dst)
    if not log_dst.is_dir():
        raise NotADirectoryError(
            f"destination for log file is not a directory: {log_dst}"
        )

    if timestamp is None:
        timestamp = timenow.get_timenow_as_str()
    logfile_name = log_dst.joinpath(f"{log_stem}_{timestamp}.log")

    if logger.hasHandlers():
        if force:
            logger.handlers.clear()
        else:
            warnings.warn(
                f"Logger already has handlers attached:\n{logger.handlers}."
                f"Will not add new handlers, to avoid duplicate messages "
                f"and corrupted logs. To override, set ``force=True`` "
                f"when calling this function."
            )
            return

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler(logfile_name)
    stream_handler = logging.StreamHandler(sys.stdout)
    for handler in (file_handler, stream_handler):
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(level)


def log_version(logger: logging.Logger) -> None:
    logger.info(
            f"vak version: {__version__}"
    )



"""tests for vak.cli.train module"""
from unittest import mock

import pytest

import vak.config
import vak.constants
import vak.paths
import vak.cli.train

from . import cli_asserts


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_train(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    root_results_dir = tmp_path.joinpath("test_train_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "TRAIN",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "TRAIN", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="train",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    with mock.patch('vak.core.train', autospec=True) as mock_core_train:
        vak.cli.train.train(toml_path)
        assert mock_core_train.called

    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]
    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="train", output_path=results_path)
    assert cli_asserts.log_file_contains_version(command="train", output_path=results_path)


def test_train_csv_path_none_raises(
        specific_config, tmp_path,
):
    """Test that cli.train raises ValueError when csv_path is None
    (presumably because `vak prep` was not run yet)
    """
    root_results_dir = tmp_path.joinpath("test_train_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {"section": "TRAIN", "option": "root_results_dir", "value": str(root_results_dir)},
        {"section": "TRAIN", "option": "csv_path", "value": "DELETE-OPTION"},
    ]

    toml_path = specific_config(
        config_type="train",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )

    with pytest.raises(ValueError):
        vak.cli.train.train(toml_path)


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("train", "cbin", None, "notmat"),
        ("train", "wav", None, "birdsong-recognition-dataset"),
        ("train", None, "mat", "yarden"),
        ("train_continue", "cbin", None, "notmat"),
        ("train_continue", "wav", None, "birdsong-recognition-dataset"),
        ("train_continue", None, "mat", "yarden"),
    ],
)
def test_train_passes_correct_labelset_and_labelmap_path(
    config_type, audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    """Test that ``cli.train`` passes in the expected arguments to ``core.train``,
    depending on whether ``labelmap_path`` is specified in the [TRAIN] section
    of the config file or not.
    """
    root_results_dir = tmp_path.joinpath("test_train_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {"section": "TRAIN", "option": "root_results_dir", "value": str(root_results_dir)},
        {"section": "TRAIN", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type=config_type,
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    cfg = vak.config.parse.from_toml_path(toml_path)

    with mock.patch('vak.core.train', autospec=True) as mock_core_train:
        vak.cli.train.train(toml_path)
        if config_type == "train":
            assert mock_core_train.call_args[1]['labelset'] == cfg.prep.labelset
            assert mock_core_train.call_args[1]['labelmap_path'] is None
        elif config_type == "train_continue":
            assert mock_core_train.call_args[1]['labelset'] is None
            assert mock_core_train.call_args[1]['labelmap_path'] == cfg.train.labelmap_path



"""script to download sample data for vak autoannotate tutorial

Adapted from
https://github.com/NickleDave/bfsongrepo/blob/main/src/scripts/download_dataset.py
"""
from __future__ import annotations
import argparse
import pathlib
import shutil
import sys
import time
from typing import Union
import urllib.request
import warnings


DATA_TO_DOWNLOAD = {
    "gy6or6": {
        "sober.repo1.gy6or6.032212.wav.csv.tar.gz": {
            "MD5": "8c88b46ba87f9784d3690cc8ee4bf2f4",
            "download": "https://figshare.com/ndownloader/files/37509160"
        },
        "sober.repo1.gy6or6.032312.wav.csv.tar.gz": {
            "MD5": "063ba4d50d1b94009b4b00f0a941d098",
            "download": "https://figshare.com/ndownloader/files/37509172"
        }
    }
}


def reporthook(count: int, block_size: int, total_size: int) -> None:
    """hook for urlretrieve that gives us a simple progress report
    https://blog.shichao.io/2012/10/04/progress_speed_indicator_for_urlretrieve_in_python.html
    """
    global start_time
    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = int(progress_size / (1024 * duration))
    percent = int(count * block_size * 100 / total_size)
    sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed" %
                    (percent, progress_size / (1024 * 1024), speed, duration))
    sys.stdout.flush()


def download_dataset(download_urls_by_bird_ID: dict,
                     bfsongrepo_dir: pathlib.Path) -> None:
    """download the dataset, given a dict of download urls"""
    tar_dir = bfsongrepo_dir / "tars"
    tar_dir.mkdir()
    # top-level keys are bird ID: bl26lb16, gr41rd51, ...
    for bird_id, tars_dict in download_urls_by_bird_ID.items():
        print(
            f'Downloading .tar files for bird: {bird_id}'
        )
        # bird ID -> dict where keys are .tar.gz filenames mapping to download url + MD5 hash
        for tar_name, url_md5_dict in tars_dict.items():
            print(
                f'Downloading tar: {tar_name}'
            )
            download_url = url_md5_dict['download']
            filename = tar_dir / tar_name
            urllib.request.urlretrieve(download_url, filename, reporthook)
            print('\n')


def extract_tars(bfsongrepo_dir: pathlib.Path) -> None:
    tar_dir = bfsongrepo_dir / "tars"  # made by download_dataset function
    tars = sorted(tar_dir.glob('*.tar.gz'))
    for tar_path in tars:
        print(
            f"\nunpacking: {tar_path}"
        )

        shutil.unpack_archive(
            filename=tar_path,
            extract_dir=bfsongrepo_dir,
            format="gztar"
        )


def main(dst: Union[str, pathlib.Path]) -> None:
    """main function that downloads and extracts entire dataset"""
    dst = pathlib.Path(dst).expanduser().resolve()
    if not dst.is_dir():
        raise NotADirectoryError(
            f"Value for 'dst' argument not recognized as a directory: {dst}"
        )
    bfsongrepo_dir = dst / 'bfsongrepo'
    if bfsongrepo_dir.exists():
        warnings.warn(
            f"Directory already exists: {bfsongrepo_dir}\n"
            "Will download and write over any existing files. Press Ctrl-C to stop."
        )

    try:
        bfsongrepo_dir.mkdir(exist_ok=True)
    except PermissionError as e:
        raise PermissionError(
            f"Unable to create directory in 'dst': {dst}\n"
            "Please try running with 'sudo' on Unix systems or as Administrator on Windows systems.\n"
            "If that fails, please download files for tutorial manually from the 'download' links in tutorial page."
        ) from e

    print(
        f'Downloading Bengalese Finch Song Repository to: {bfsongrepo_dir}'
    )

    download_dataset(DATA_TO_DOWNLOAD, bfsongrepo_dir)
    extract_tars(bfsongrepo_dir)


def get_parser() -> argparse.ArgumentParser:
    """get ArgumentParser used to parse command-line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dst',
        default='.',
        help=("Destination where dataset should be downloaded. "
              "Default is '.', i.e., current working directory "
              "from which this script is run.'")
    )
    return parser


parser = get_parser()
args = parser.parse_args()
main(dst=args.dst)



"""tests for vak.core.eval module"""
import pytest

import vak.config
import vak.constants
import vak.paths
import vak.core.eval


# written as separate function so we can re-use in tests/unit/test_cli/test_eval.py
def eval_output_matches_expected(model_config_map, output_dir):
    for model_name in model_config_map.keys():
        eval_csv = sorted(output_dir.glob(f"eval_{model_name}*csv"))
        assert len(eval_csv) == 1

    return True


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
    ],
)
def test_eval(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "EVAL", "option": "output_dir", "value": str(output_dir)},
        {"section": "EVAL", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="eval",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.eval.models)

    vak.core.eval(
        cfg.eval.csv_path,
        model_config_map,
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        window_size=cfg.dataloader.window_size,
        num_workers=cfg.eval.num_workers,
        spect_scaler_path=cfg.eval.spect_scaler_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        device=cfg.eval.device,
    )

    assert eval_output_matches_expected(model_config_map, output_dir)


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"section": "EVAL", "option": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
        {"section": "EVAL", "option": "labelmap_path", "value": '/obviously/doesnt/exist/labelmap.json'},
        {"section": "EVAL", "option": "csv_path", "value": '/obviously/doesnt/exist/dataset.csv'},
        {"section": "EVAL", "option": "spect_scaler_path", "value": '/obviously/doesnt/exist/SpectScaler'},
    ]
)
def test_eval_raises_file_not_found(
    path_option_to_change,
    specific_config,
    tmp_path,
    device
):
    """Test that core.eval raises FileNotFoundError
    when one of the following does not exist:
    checkpoint_path, labelmap_path, csv_path, spect_scaler_path
    """
    output_dir = tmp_path.joinpath(
        f"test_eval_cbin_notmat_invalid_csv_path"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "EVAL", "option": "output_dir", "value": str(output_dir)},
        {"section": "EVAL", "option": "device", "value": device},
        path_option_to_change,
    ]

    toml_path = specific_config(
        config_type="eval",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.eval.models)
    with pytest.raises(FileNotFoundError):
        vak.core.eval(
            csv_path=cfg.eval.csv_path,
            model_config_map=model_config_map,
            checkpoint_path=cfg.eval.checkpoint_path,
            labelmap_path=cfg.eval.labelmap_path,
            output_dir=cfg.eval.output_dir,
            window_size=cfg.dataloader.window_size,
            num_workers=cfg.eval.num_workers,
            spect_scaler_path=cfg.eval.spect_scaler_path,
            spect_key=cfg.spect_params.spect_key,
            timebins_key=cfg.spect_params.timebins_key,
            device=cfg.eval.device,
        )


def test_eval_raises_not_a_directory(
    specific_config,
    device
):
    """Test that core.eval raises NotADirectory
    when ``output_dir`` does not exist
    """
    options_to_change = [
        {"section": "EVAL", "option": "output_dir", "value": '/obviously/does/not/exist/output'},
        {"section": "EVAL", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="eval",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.eval.models)
    with pytest.raises(NotADirectoryError):
        vak.core.eval(
            csv_path=cfg.eval.csv_path,
            model_config_map=model_config_map,
            checkpoint_path=cfg.eval.checkpoint_path,
            labelmap_path=cfg.eval.labelmap_path,
            output_dir=cfg.eval.output_dir,
            window_size=cfg.dataloader.window_size,
            num_workers=cfg.eval.num_workers,
            spect_scaler_path=cfg.eval.spect_scaler_path,
            spect_key=cfg.spect_params.spect_key,
            timebins_key=cfg.spect_params.timebins_key,
            device=cfg.eval.device,
        )



import pathlib
from unittest import mock

import pytest

import vak


@pytest.fixture
def parser():
    return vak.__main__.get_parser()


def test_parser_usage(parser,
                      capsys):
    with pytest.raises(SystemExit):
        parser.parse_args(args=[''])
    captured = capsys.readouterr()
    assert captured.err.startswith(
        "usage: vak [-h] command configfile"
    )


def test_parser_help(parser,
                     capsys):
    with pytest.raises(SystemExit):
        parser.parse_args(['-h'])
    captured = capsys.readouterr()
    assert captured.out.startswith(
        "usage: vak [-h] command configfile"
    )


DUMMY_CONFIGFILE = './configs/config_2018-12-17.toml'


@pytest.mark.parametrize(
    'command, raises',
    [
        ('prep', False),
        ('train', False),
        ('learncurve', False),
        ('eval', False),
        ('predict', False),
        ('not-a-valid-command', True),
    ]
)
def test_parser(command,
                raises,
                parser,
                capsys):
    if raises:
        with pytest.raises(SystemExit):
            parser.parse_args([command, DUMMY_CONFIGFILE])
    else:
        args = parser.parse_args([command, DUMMY_CONFIGFILE])
        assert args.command == command
        assert args.configfile == pathlib.Path(DUMMY_CONFIGFILE)


@pytest.mark.parametrize(
    'command',
    [
        'prep',
        'train',
        'learncurve',
        'eval',
        'predict',
    ]
)
def test_main(command,
              parser):
    args = parser.parse_args([command, DUMMY_CONFIGFILE])
    mock_cli_function = mock.Mock(name=f'mock_{command}')
    with mock.patch.dict(vak.cli.cli.COMMAND_FUNCTION_MAP,
                         {command: mock_cli_function}) as mock_command_function_map:
        vak.__main__.main(args)
        mock_cli_function.assert_called()



import os
import pathlib
import shutil
import tarfile
import urllib.request

import nox


DIR = pathlib.Path(__file__).parent.resolve()
VENV_DIR = pathlib.Path('./.venv').resolve()

nox.options.sessions = ['test', 'coverage']


@nox.session
def build(session: nox.Session) -> None:
    """
    Build an SDist and wheel with ``flit``.
    """
    dist_dir = DIR.joinpath("dist")
    if dist_dir.exists():
        shutil.rmtree(dist_dir)

    session.install(".[dev]")
    session.run("flit", "build")


@nox.session
def dev(session: nox.Session) -> None:
    """
    Sets up a python development environment for the project.

    This session will:
    - Create a python virtualenv for the session
    - Install the `virtualenv` cli tool into this environment
    - Use `virtualenv` to create a global project virtual environment
    - Invoke the python interpreter from the global project environment to install
      the project and all it's development dependencies.
    """
    session.install("virtualenv")
    # VENV_DIR here is a pathlib.Path location of the project virtualenv
    # e.g. .venv
    session.run("virtualenv", os.fsdecode(VENV_DIR), silent=True)

    python = os.fsdecode(VENV_DIR.joinpath("bin/python"))

    # Use the venv's interpreter to install the project along with
    # all it's dev dependencies, this ensures it's installed in the right way
    session.run(python, "-m", "pip", "install", "-e", ".[dev,test,doc]", external=True)


# ---- used by sessions that "clean up" data for tests
def clean_dir(dir_path):
    """
    "clean" a directory by removing all files
    (that are not hidden)
    without removing the directory itself
    """
    dir_path = pathlib.Path(dir_path)
    dir_contents = dir_path.glob('*')
    for content in dir_contents:
        if content.is_dir():
            shutil.rmtree(content)
        else:
            if content.name.startswith('.'):
                # e.g., .gitkeep file we don't want to delete
                continue
            content.unlink()


DATA_FOR_TESTS_DIR = './tests/data_for_tests/'
SOURCE_TEST_DATA_DIR = f"{DATA_FOR_TESTS_DIR}source/"
SOURCE_TEST_DATA_DIRS = [
    dir_ for dir_
    in sorted(pathlib.Path(SOURCE_TEST_DATA_DIR).glob('*/'))
    if dir_.is_dir()
]


@nox.session(name='test-data-clean-source')
def test_data_clean_source(session) -> None:
    """
    Clean (remove) 'source' test data, used by TEST_DATA_GENERATE_SCRIPT.
    """
    clean_dir(SOURCE_TEST_DATA_DIR)


def copy_url(url: str, path: str) -> None:
    """Copy data from a url to a local file."""
    urllib.request.urlretrieve(url, path)


SOURCE_TEST_DATA_URL = 'https://osf.io/hbg4k/download'
SOURCE_TEST_DATA_TAR = f'{SOURCE_TEST_DATA_DIR}source_test_data.tar.gz'


@nox.session(name='test-data-tar-source')
def test_data_tar_source(session) -> None:
    """
    Make a .tar.gz file of just the 'generated' test data used to run tests on CI.
    """
    session.log(f"Making tarfile with source data: {SOURCE_TEST_DATA_TAR}")
    make_tarfile(SOURCE_TEST_DATA_TAR, SOURCE_TEST_DATA_DIRS)


@nox.session(name='test-data-download-source')
def test_data_download_source(session) -> None:
    """
    Download and extract a .tar.gz file of 'source' test data, used by TEST_DATA_GENERATE_SCRIPT.
    """
    session.log(f'Downloading: {SOURCE_TEST_DATA_URL}')
    copy_url(url=SOURCE_TEST_DATA_URL, path=SOURCE_TEST_DATA_TAR)
    session.log(f'Extracting downloaded tar: {SOURCE_TEST_DATA_TAR}')
    with tarfile.open(SOURCE_TEST_DATA_TAR, "r:gz") as tf:
        tf.extractall(path='.')


TEST_DATA_GENERATE_SCRIPT = './tests/scripts/generate_data_for_tests.py'


@nox.session(name='test-data-generate')
def test_data_generate(session) -> None:
    """
    Produced 'generated' test data, by running TEST_DATA_GENERATE_SCRIPT on 'source' test data.
    """
    session.install(".[test]")
    session.run("python", TEST_DATA_GENERATE_SCRIPT)


GENERATED_TEST_DATA_DIR = f'{DATA_FOR_TESTS_DIR}generated/'


@nox.session(name='test-data-clean-generated')
def test_data_clean_generated(session) -> None:
    """
    Clean (remove) 'generated' test data.
    """
    clean_dir(GENERATED_TEST_DATA_DIR)


def make_tarfile(name: str, to_add: list):
    with tarfile.open(name, "w:gz") as tf:
        for add_name in to_add:
            tf.add(name=add_name)


CONFIGS_DIR = f'{GENERATED_TEST_DATA_DIR}configs'
PREP_DIR = f'{GENERATED_TEST_DATA_DIR}prep/'
RESULTS_DIR = f'{GENERATED_TEST_DATA_DIR}results/'

PREP_CI = sorted(pathlib.Path(PREP_DIR).glob('*/*/teenytweetynet'))
RESULTS_CI = sorted(pathlib.Path(RESULTS_DIR).glob('*/*/teenytweetynet'))
GENERATED_TEST_DATA_CI_TAR = f'{GENERATED_TEST_DATA_DIR}generated_test_data.ci.tar.gz'
GENERATED_TEST_DATA_CI_DIRS = [CONFIGS_DIR] + PREP_CI + RESULTS_CI

GENERATED_TEST_DATA_ALL_TAR = f'{GENERATED_TEST_DATA_DIR}generated_test_data.tar.gz'
GENERATED_TEST_DATA_ALL_DIRS = [CONFIGS_DIR, PREP_DIR, RESULTS_DIR]


@nox.session(name='test-data-tar-generated-all')
def test_data_tar_generated_all(session) -> None:
    """
    Make a .tar.gz file of all 'generated' test data.
    """
    session.log(f"Making tarfile with all generated data: {GENERATED_TEST_DATA_ALL_TAR}")
    make_tarfile(GENERATED_TEST_DATA_ALL_TAR, GENERATED_TEST_DATA_ALL_DIRS)


@nox.session(name='test-data-tar-generated-ci')
def test_data_tar_generated_ci(session) -> None:
    """
    Make a .tar.gz file of just the 'generated' test data used to run tests on CI.
    """
    session.log(f"Making tarfile with generated data for CI: {GENERATED_TEST_DATA_CI_TAR}")
    make_tarfile(GENERATED_TEST_DATA_CI_TAR, GENERATED_TEST_DATA_CI_DIRS)


GENERATED_TEST_DATA_ALL_URL = 'https://osf.io/9qmb4/download'


@nox.session(name='test-data-download-generated-all')
def test_data_download_generated_all(session) -> None:
    """
    Download and extract a .tar.gz file of all 'generated' test data
    """
    session.log(f'Downloading: {GENERATED_TEST_DATA_ALL_URL}')
    copy_url(url=GENERATED_TEST_DATA_ALL_URL, path=GENERATED_TEST_DATA_ALL_TAR)
    session.log(f'Extracting downloaded tar: {GENERATED_TEST_DATA_ALL_TAR}')
    with tarfile.open(GENERATED_TEST_DATA_ALL_TAR, "r:gz") as tf:
        tf.extractall(path='.')
    session.log('Fixing paths in .csv files')
    session.run(
        "python", "./tests/scripts/fix_prep_csv_paths.py"
    )


GENERATED_TEST_DATA_CI_URL = 'https://osf.io/36gwp/download'


@nox.session(name='test-data-download-generated-ci')
def test_data_download_generated_ci(session) -> None:
    """
    Download and extract a .tar.gz file of just the 'generated' test data used to run tests on CI
    """
    session.install("pandas")
    session.log(f'Downloading: {GENERATED_TEST_DATA_CI_URL}')
    copy_url(url=GENERATED_TEST_DATA_CI_URL, path=GENERATED_TEST_DATA_CI_TAR)
    session.log(f'Extracting downloaded tar: {GENERATED_TEST_DATA_CI_TAR}')
    with tarfile.open(GENERATED_TEST_DATA_CI_TAR, "r:gz") as tf:
        tf.extractall(path='.')
    session.log('Fixing paths in .csv files')
    session.run(
        "python", "./tests/scripts/fix_prep_csv_paths.py"
    )


@nox.session
def test(session) -> None:
    """
    Run the unit and regular tests.
    """
    session.install(".[test]")
    session.run("pytest", *session.posargs)


@nox.session
def coverage(session) -> None:
    """
    Run the unit and regular tests, and save coverage report
    """
    session.install(".[test]")
    if session.posargs:
        if "running-on-ci" in session.posargs:
            # on ci, just run `teenytweetynet` model
            session.run(
                "pytest", "--models", "teenytweetynet", "--cov=./", "--cov-report=xml"
            )
            return
        else:
            print("Unsupported argument to coverage")

    session.run(
        "pytest", "--cov=./", "--cov-report=xml", *session.posargs
    )


@nox.session
def doc(session: nox.Session) -> None:
    """
    Build the docs.

    To run ``sphinx-autobuild``,  do:

    .. code-block::console

       nox -s doc -- autobuild

    Otherwise the docs will be built once using
    """
    session.install(".[doc]")
    if session.posargs:
        if "autobuild" in session.posargs:
            print("Building docs at http://127.0.0.1:8000 with sphinx-autobuild -- use Ctrl-C to quit")
            session.run("sphinx-autobuild", "doc", "doc/_build/html")
        else:
            print("Unsupported argument to docs")
    else:
        session.run("sphinx-build", "-nW", "--keep-going", "-b", "html", "doc/", "doc/_build/html")



"""tests for vak.cli.learncurve module"""
from unittest import mock

import pytest

import vak.config
import vak.constants
import vak.cli.learncurve

from . import cli_asserts


def test_learncurve(specific_config, tmp_path, model, device):
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    with mock.patch('vak.core.learning_curve', autospec=True) as mock_core_learning_curve:
        vak.cli.learncurve.learning_curve(toml_path)
        assert mock_core_learning_curve.called

    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="learncurve", output_path=results_path)
    assert cli_asserts.log_file_contains_version(command="learncurve", output_path=results_path)


@pytest.mark.parametrize("window_size", [44, 88, 176])
def test_learncurve_previous_run_path(
    specific_config, tmp_path, model, device, previous_run_path_factory, window_size
):
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {
            "section": "LEARNCURVE",
            "option": "device",
            "value": device},
        {
            "section": "LEARNCURVE",
            "option": "previous_run_path",
            "value": str(previous_run_path_factory(model)),
        },
        {"section": "DATALOADER", "option": "window_size", "value": window_size},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    with mock.patch('vak.core.learning_curve', autospec=True) as mock_core_learning_curve:
        vak.cli.learncurve.learning_curve(toml_path)
        assert mock_core_learning_curve.called

    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="learncurve", output_path=results_path)
    assert cli_asserts.log_file_contains_version(command="learncurve", output_path=results_path)


def test_learning_curve_csv_path_none_raises(
        specific_config, tmp_path,
):
    """Test that cli.learncurve.learning_curve
    raises ValueError when csv_path is None
    (presumably because `vak prep` was not run yet)
    """
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {
            "section": "LEARNCURVE",
            "option": "csv_path",
            "value": "DELETE-OPTION"},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model="teenytweetynet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        options_to_change=options_to_change,
    )

    with pytest.raises(ValueError):
        vak.cli.learncurve.learning_curve(toml_path)



import logging
from pathlib import Path

from .. import (
    config,
    core,
    validators
)
from ..logging import config_logging_for_cli, log_version


logger = logging.getLogger(__name__)


def eval(toml_path):
    """evaluate a trained model with dataset specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml_path(toml_path)

    if cfg.eval is None:
        raise ValueError(
            f"eval called with a config.toml file that does not have a EVAL section: {toml_path}"
        )

    # ---- set up logging ---------------------------------------------------------------------------------------------
    config_logging_for_cli(
        log_dst=cfg.eval.output_dir,
        log_stem="eval",
        level="INFO",
        force=True
    )
    log_version(logger)

    logger.info("Logging results to {}".format(cfg.eval.output_dir))

    model_config_map = config.models.map_from_path(toml_path, cfg.eval.models)

    if cfg.eval.csv_path is None:
        raise ValueError(
            "No value is specified for 'csv_path' in this .toml config file."
            f"To generate a .csv file that represents the dataset, "
            f"please run the following command:\n'vak prep {toml_path}'"
        )

    core.eval(
        cfg.eval.csv_path,
        model_config_map,
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        window_size=cfg.dataloader.window_size,
        num_workers=cfg.eval.num_workers,
        spect_scaler_path=cfg.eval.spect_scaler_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        device=cfg.eval.device,
    )



import logging
from pathlib import Path
import shutil

from .. import (
    config,
    core
)
from ..logging import config_logging_for_cli, log_version
from ..paths import generate_results_dir_name_as_path


logger = logging.getLogger(__name__)


def learning_curve(toml_path):
    """generate learning curve, by training models on training sets across a
    range of sizes and then measure accuracy of those models on a test set.
    Function called by command-line interface.

    Trains models, saves results in new directory within root_results_dir specified
    in config.toml file, and adds path to that new directory to config.toml file.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml_path(toml_path)

    if cfg.learncurve is None:
        raise ValueError(
            f"train called with a config.toml file that does not have a TRAIN section: {toml_path}"
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    results_path = generate_results_dir_name_as_path(cfg.learncurve.root_results_dir)
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    config_logging_for_cli(
        log_dst=results_path,
        log_stem="learncurve",
        level="INFO",
        force=True
    )
    log_version(logger)
    logger.info("Logging results to {}".format(results_path))

    model_config_map = config.models.map_from_path(toml_path, cfg.learncurve.models)

    if cfg.learncurve.csv_path is None:
        raise ValueError(
            "No value is specified for 'csv_path' in this .toml config file."
            f"To generate a .csv file that represents the dataset, "
            f"please run the following command:\n'vak prep {toml_path}'"
        )

    core.learning_curve(
        model_config_map,
        train_set_durs=cfg.learncurve.train_set_durs,
        num_replicates=cfg.learncurve.num_replicates,
        csv_path=cfg.learncurve.csv_path,
        labelset=cfg.prep.labelset,
        window_size=cfg.dataloader.window_size,
        batch_size=cfg.learncurve.batch_size,
        num_epochs=cfg.learncurve.num_epochs,
        num_workers=cfg.learncurve.num_workers,
        results_path=results_path,
        previous_run_path=cfg.learncurve.previous_run_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        normalize_spectrograms=cfg.learncurve.normalize_spectrograms,
        shuffle=cfg.learncurve.shuffle,
        val_step=cfg.learncurve.val_step,
        ckpt_step=cfg.learncurve.ckpt_step,
        patience=cfg.learncurve.patience,
        device=cfg.learncurve.device,
    )
